local S = minetest.get_translator("lzr_level_select")
local F = minetest.formspec_escape

lzr_level_select = {}

local current_level_selection = nil
local current_custom_level_selection = nil
local custom_levels

local message_form = function(player, message)
	return "formspec_version[7]size[8,3.5]"..
	"textarea[0.5,0.5;7,1.8;;;"..F(message).."]"..
	"button_exit[3,2.5;2,0.6;okay;"..F(S("OK")).."]"
end

lzr_level_select.open_dialog = function(player, level_set)
	local caption
	if level_set == "core" then
		caption = S("Select level:")
	elseif level_set == "custom" then
		caption = S("Select custom level:")
	else
		minetest.log("error", "[lzr_level_select] open_dialog called with unknown level_set!")
		return false
	end
	local form = "formspec_version[7]size[6,10]"..
		"label[0.5,0.4;"..F(caption).."]"..
		"button_exit[1.5,8.5;3,1;okay;"..F(S("Start")).."]"..
		"tablecolumns[color;text]"..
		"table[0.5,0.8;5,7.5;levellist;"
	local list = {}
	local entry_header = ""
	local first_uncompleted_level = nil
	if level_set == "core" then
		-- Built-in level set
		local completed_levels = lzr_levels.get_completed_levels()
		local level_data = lzr_levels.get_core_level_data()
		for l=1, #level_data do
			local filename = level_data[l].filename
			local level_id = string.sub(filename, 1, -5)
			if completed_levels[level_id] then
				entry_header = "#00FF00"
			else
				if not first_uncompleted_level then
					first_uncompleted_level = l
				end
				entry_header  = ""
			end
			table.insert(list, entry_header..","..F(lzr_levels.get_level_name(l, nil, true)))
		end
		if not first_uncompleted_level then
			first_uncompleted_level = 1
		end
		current_level_selection = first_uncompleted_level
	else
		-- Custom level set (editor levels)
		local list_levels = lzr_editor.get_custom_levels()
		custom_levels = {}
		for l=1, #list_levels do
			local levelname = list_levels[l]
			-- Hide the autosave level because it's not meant
			-- for playing
			if levelname ~= lzr_globals.AUTOSAVE_NAME then
				table.insert(custom_levels, levelname)

				local proper_name = lzr_editor.get_custom_level_name(levelname, true)
				table.insert(list, entry_header..","..F(proper_name))
			end
		end
		first_uncompleted_level = 1
		current_custom_level_selection = 1
	end
	if not first_uncompleted_level then
		first_uncompleted_level = 1
	end
	local list_str = table.concat(list, ",")
	form = form .. list_str .. ";"..first_uncompleted_level.."]"
	local formname
	if level_set == "core" then
		if #list == 0 then
			form = message_form(player, S("There are no levels."))
		end
		formname = "lzr_level_select:levellist"
	else
		if #list == 0 then
			form = message_form(player,
				S("There are no custom levels.").."\n"..
				S("Install levels from your friends or use the level editor to build your own."))
		end
		formname = "lzr_level_select:levellist_custom"
	end
	minetest.show_formspec(player:get_player_name(), formname, form)
end

local load_custom_level = function(level, player)
	minetest.log("action", "[lzr_level_select] Player selects custom level: "..level)
	local csv = minetest.get_worldpath().."/levels/"..level..".csv"
	local levels_path = minetest.get_worldpath().."/levels"
	local pname = player:get_player_name()

	local custom_level_data, error_type, error_msg, error_detail = lzr_levels.analyze_levels(csv, levels_path)
	if custom_level_data then
		-- Load level with metadata (preferred)
		lzr_levels.start_level(1, custom_level_data)
	elseif error_type == "load_error" then
		-- Fallback when CSV file could not be read
		custom_level_data = lzr_levels.create_fallback_level_data(level, levels_path)
		if not custom_level_data then
			local form = message_form(player, S("Level could not be loaded. No level CSV file was found, and the fallback mechanism failed as well."))
			minetest.show_formspec(pname, "lzr_level_select:custom_load_error", form)
			minetest.log("action", "[lzr_level_select] Player tried to load broken custom level: "..tostring(level)..". Missing CSV file, and the fallback, failed as well.")
		else
			lzr_levels.start_level(1, custom_level_data)
			minetest.log("action", "[lzr_level_select] No CSV file found for custom level: "..tostring(level)..". Using fallback.")
		end
	elseif error_type == "csv_error" then
		local form = message_form(player, S("The level could not be loaded. Invalid syntax of level CSV file."))
		minetest.show_formspec(pname, "lzr_level_select:custom_load_error", form)
		minetest.log("action", "[lzr_level_select] Player tried to load broken custom level "..tostring(level)..". Invalid CSV syntax.")
	elseif error_type == "bad_schematic" then
		local error_append = ""
		if error_detail then
			local reason
			if error_detail == "no_teleporter" then
				reason = S("There’s no teleporter for the player to start on.")
			elseif error_detail == "too_many_teleporters" then
				reason = S("There’s more than one teleporter.")
			elseif error_detail == "too_many_parrot_spawners" then
				reason = S("There’s more than one information block.")
			elseif error_detail == "barriers" then
				reason = S("There’s a barrier or barrier-like node in the level.")
			elseif error_detail == "gold_block" then
				reason = S("There’s a bare gold block in the level.")
			elseif error_detail == "plant_on_ground" then
				reason = S("There’s a rooted plant in the level.")
			else
				reason = error_detail
			end
			error_append = "\n\n" .. S("Reason: @1", reason)
		end
		local form = message_form(player, S("This level is unplayable.")..error_append)
		minetest.show_formspec(pname, "lzr_level_select:custom_load_error", form)
		minetest.log("action", "[lzr_level_select] Player tried to load broken custom level "..tostring(level)..". Bad schematic.")
	else
		local form = message_form(player, S("The level could not be loaded due to an unknown error."))
		minetest.show_formspec(pname, "lzr_level_select:custom_load_error", form)
		minetest.log("error", "[lzr_level_select] Unknown error while loading custom level: "..tostring(level)..". Aborting.")
	end
end

minetest.register_on_player_receive_fields(function(player, formname, fields)
	if lzr_gamestate.get_state() ~= lzr_gamestate.MENU then
		return
	end
	if formname == "lzr_level_select:levellist" then
		if fields.okay then
			if current_level_selection then
				lzr_levels.start_level(current_level_selection)
			end
		elseif fields.levellist then
			local expl = minetest.explode_table_event(fields.levellist)
			if expl.type == "CHG" then
				current_level_selection = expl.row
			elseif expl.type == "DCL" then
				current_level_selection = expl.row
				lzr_levels.start_level(current_level_selection)
				minetest.close_formspec(player:get_player_name(), "lzr_level_select:levellist")
			elseif expl.type == "INV" then
				current_level_selection = nil
			end
		end
	elseif formname == "lzr_level_select:levellist_custom" then
		if fields.okay then
			if custom_levels and current_custom_level_selection then
				local level = custom_levels[current_custom_level_selection]
				if level then
					load_custom_level(level, player)
				end
			end
		elseif fields.levellist then
			local expl = minetest.explode_table_event(fields.levellist)
			if expl.type == "CHG" then
				current_custom_level_selection = expl.row
			elseif expl.type == "DCL" then
				current_custom_level_selection = expl.row
				if custom_levels then
					local level = custom_levels[current_custom_level_selection]
					if level then
						load_custom_level(level, player)
					end
				end
				minetest.close_formspec(player:get_player_name(), "lzr_level_select:levellist_custom")
			elseif expl.type == "INV" then
				current_custom_level_selection = nil
			end
		end
	end
end)

minetest.register_chatcommand("level", {
	privs = { server = true },
	description = S("Go to level"),
	params = S("<level>"),
	func = function(name, param)
		local level = tonumber(param)
		if not level then
			return false
		end
		if level < 1 or level > lzr_levels.LAST_LEVEL then
			return false, S("Invalid level!")
		end
		if lzr_gamestate.is_loading() then
			return false, S("Can’t start a level while loading!")
		end
		lzr_levels.start_level(level)
		return true
	end,

})

